<?php

namespace Dotorg\API\Events;

if ( 'cli' !== php_sapi_name() ) {
	die();
}

/**
 * Main entry point
 */
function run_tests() {
	global $wpdb;
	define( 'RUNNING_TESTS', true );
	define( 'SAVEQUERIES',   true );

	require_once( dirname( __DIR__ ) . '/index.php' );

	$tests_failed = 0;
	$tests_failed += test_get_location();
	$query_count  = count( $wpdb->queries );
	$query_time   = array_sum( array_column( $wpdb->queries, 1 ) );

	printf(
		"\n\nFinished running all tests.\n\n* %d tests failed\n* %d queries ran in %f seconds\n* Average time per query: %f seconds\n",
		$tests_failed,
		$query_count,
		$query_time,
		$query_time / $query_count
	);
}

/**
 * Output the results of an individual test
 *
 * @param int   $case_id
 * @param bool  $passed
 * @param mixed $expected_result
 * @param mixed $actual_result
 */
function output_results( $case_id, $passed, $expected_result, $actual_result ) {
	printf(
		"\n* %s: %s",
		$case_id,
		$passed ? 'PASSED' : '_FAILED_'
	);

	if ( ! $passed ) {
		$expected_output = is_scalar( $expected_result ) ? var_export( $expected_result, true ) : print_r( $expected_result, true );
		$actual_output   = is_scalar( $actual_result   ) ? var_export( $actual_result,   true ) : print_r( $actual_result,   true );

		printf(
			"\n\nExpected result: %s\nActual result: %s",
			$expected_output,
			$actual_output
		);
	}
}

/**
 * Add a cachebusting parameter to bypass the object cache
 *
 * Cache keys are generated based on the function's input arguments (e.g., get_location()), so adding a unique
 * parameter on every function call ensures that the unit tests will never get a cached result.
 *
 * @param array $arguments
 *
 * @return array
 */
function add_cachebusting_parameter( $arguments ) {
	$arguments['cachebuster'] = microtime( true );

	return $arguments;
}

/**
 * Test `get_location()`
 *
 * @return bool The number of failures
 */
function test_get_location() {
	$failed = 0;
	$cases  = get_location_test_cases();

	printf( "\nRunning %d location tests\n", count( $cases ) );

	foreach ( $cases as $case_id => $case ) {
		$case['input'] = add_cachebusting_parameter( $case['input'] );
		$actual_result = get_location( $case['input'] );

		// Normalize to lowercase to account for inconsistency in the IP database
		if ( isset( $actual_result['description'] ) && is_string( $actual_result['description'] ) ) {
			$actual_result['description'] = strtolower( $actual_result['description'] );
		}

		/*
		 * Normalize coordinates to account for minor differences in the databases
		 *
		 * Rounding to three decimal places means that we're still accurate within about 110 meters, which is
		 * good enough for our purposes.
		 *
		 * See https://gis.stackexchange.com/a/8674/49125
		 */
		if ( isset( $actual_result['latitude'], $actual_result['longitude'] ) ) {
			$actual_result['latitude']  = number_format( round( $actual_result['latitude'],  3 ), 3 );
			$actual_result['longitude'] = number_format( round( $actual_result['longitude'], 3 ), 3 );
		}

		$passed      = $case['expected'] === $actual_result;

		output_results( $case_id, $passed, $case['expected'], $actual_result );

		if ( ! $passed ) {
			$failed++;
		}
	}

	return $failed;
}

/**
 * Get the cases for testing `get_location()`
 *
 * @return array
 */
function get_location_test_cases() {
	 $cases = array(
		/*
		 * Only the country code is given
		 */
		'country-code-australia' => array(
			'input' => array(
				'country' => 'AU',
			),
			'expected' => array(
				'country' => 'AU',
			),
		),


		/*
		 * The country name, locale, and timezone are given
		 */
		'country-exonym-1-word' => array(
			'input' => array(
				'location_name' => 'Indonesia',
				'locale'        => 'id_ID',
				'timezone'      => 'Asia/Jakarta',
			),
			'expected' => array(
				'country' => 'ID',
				'description' => 'indonesia',
			),
		),

		/*
		 * This is matching a city inside the country before it the country searches run, but that's ok since it's
		 * good enough for our use cases
		 */
		'country-exonym-2-words' => array(
			'input' => array(
				'location_name' => 'Bosnia and Herzegovina',
				'locale'        => 'bs_BA',
				'timezone'      => 'Europe/Sarajevo',
			),
			'expected' => array(
				'description' => 'pale',
				'latitude'    => '43.817',
				'longitude'   => '18.569',
				'country'     => 'BA'
			),
		),


		/*
		 * A location couldn't be found
		 */
		'city-invalid-private-ip' => array(
			'input' => array(
				'location_name' => 'Rivendell',
				'ip'            => '127.0.0.1'
			),
			'expected' => false,
		),

		/*
		 * No input was provided
		 */
		'input-empty' => array(
			'input'    => array(),
			'expected' => array(),
		),


		/*
		 * The English city exonym, locale, and timezone are given
		 */
		'city-africa' => array(
			'input' => array(
				'location_name' => 'Nairobi',
				'locale'        => 'en_GB',
				'timezone'      => 'Africa/Nairobi',
			),
			'expected' => array(
				'description' => 'nairobi',
				'latitude'    => '-1.283',
				'longitude'   => '36.817',
				'country'     => 'KE',
			),
		),

		'city-asia' => array(
			'input' => array(
				'location_name' => 'Tokyo',
				'locale'        => 'ja',
				'timezone'      => 'Asia/Tokyo',
			),
			'expected' => array(
				'description' => 'tokyo',
				'latitude'    => '35.690',
				'longitude'   => '139.692',
				'country'     => 'JP',
			),
		),

		'city-europe' => array(
			'input' => array(
				'location_name' => 'Berlin',
				'locale'        => 'de_DE',
				'timezone'      => 'Europe/Berlin',
			),
			'expected' => array(
				'description' => 'berlin',
				'latitude'    => '52.524',
				'longitude'   => '13.411',
				'country'     => 'DE',
			),
		),

		'city-north-america' => array(
			'input' => array(
				'location_name' => 'Vancouver',
				'locale'        => 'en_CA',
				'timezone'      => 'America/Vancouver',
			),
			'expected' => array(
				'description' => 'vancouver',
				'latitude'    => '49.250',
				'longitude'   => '-123.119',
				'country'     => 'CA',
			),
		),

		'city-oceania' => array(
			'input' => array(
				'location_name' => 'Brisbane',
				'locale'        => 'en_AU',
				'timezone'      => 'Australia/Brisbane',
			),
			'expected' => array(
				'description' => 'brisbane',
				'latitude'    => '-27.468',
				'longitude'   => '153.028',
				'country'     => 'AU',
			),
		),

		// Many users never change the default `en_US` locale in Core
		'city-oceania-with-en_US' => array(
			'input' => array(
				'location_name' => 'Sydney',
				'locale'        => 'en_US',
				'timezone'      => 'Australia/Sydney',
			),
			'expected' => array(
				'description' => 'sydney',
				'latitude'    => '-33.868',
				'longitude'   => '151.207',
				'country'     => 'AU',
			),
		),

		'city-south-america' => array(
			'input' => array(
				'location_name' => 'Sao Paulo',
				'locale'        => 'pt_BR',
				'timezone'      => 'America/Sao_Paulo',
			),
			'expected' => array(
				'description' => 'são paulo',
				'latitude'    => '-23.548',
				'longitude'   => '-46.636',
				'country'     => 'BR',
			),
		),

		// Users will often type them without the dash, bypassing an exact match
		'city-with-dashes-in-formal-name' => array(
			'input' => array(
				'location_name' => 'Osakashi',
				'locale'        => 'ja',
				'timezone'      => 'Asia/Tokyo',
			),
			'expected' => array(
				'description' => 'osaka',
				'latitude'    => '34.694',
				'longitude'   => '135.502',
				'country'     => 'JP',
			),
		),

		// If a location is provided, the fallback search should be attempted before an IP search
		'fallback-with-public-ip' => array(
			'input' => array(
				'location_name' => 'Osakashi',
				'locale'        => 'ja',
				'timezone'      => 'Asia/Tokyo',
				'ip'            => '153.163.68.148', // Tokyo
			),
			'expected' => array(
				'description' => 'osaka',
				'latitude'    => '34.694',
				'longitude'   => '135.502',
				'country'     => 'JP',
			),
		),

		'city-with-apostrophe-in-formal-name' => array(
			'input' => array(
				'location_name' => "Coeur d'Alene",
				'locale'        => 'en_US',
				'timezone'      => 'America/Los_Angeles',
			),
			'expected' => array(
				'description' => "coeur d'alene",
				'latitude'    => '47.678',
				'longitude'   => '-116.780',
				'country'     => 'US',
			),
		),

		/*
		 * todo This is failing. Not sure why at first glance, it looks like it matches the row in the database exactly.
		 *      It might be a similar issue to `Yaoundé`. If it is, this should still be kept as a separate test after
		 *      that is fixed, to make sure there aren't regressions with diacritics.
		 */
		'city-with-diacritics-in-query' => array(
			'input' => array(
				'location_name' => "Doña Ana",
				'locale'        => 'en_US',
				'timezone'      => 'America/Denver',
			),
			'expected' => array(
				'description' => "doña ana",
				'latitude'    => '32.390',
				'longitude'   => '-106.814',
				'country'     => 'US',
			),
		),

		'city-with-diacritics-in-formal-name-but-not-in-query' => array(
			'input' => array(
				'location_name' => "Dona Ana",
				'locale'        => 'en_US',
				'timezone'      => 'America/Denver',
			),
			'expected' => array(
				'description' => "doña ana",
				'latitude'    => '32.390',
				'longitude'   => '-106.814',
				'country'     => 'US',
			),
		),

		'city-with-period-in-query' => array(
			'input' => array(
				'location_name' => "St. Louis",
				'locale'        => 'en_US',
				'timezone'      => 'America/Chicago',
			),
			'expected' => array(
				'description' => "st. louis",
				'latitude'    => '38.627',
				'longitude'   => '-90.198',
				'country'     => 'US',
			),
		),

		'city-with-period-in-formal-name-but-not-in-query' => array(
			'input' => array(
				'location_name' => "St Louis",
				'locale'        => 'en_US',
				'timezone'      => 'America/Chicago',
			),
			'expected' => array(
				'description' => "st. louis",
				'latitude'    => '38.627',
				'longitude'   => '-90.198',
				'country'     => 'US',
			),
		),

		/*
		 * The city endonym, locale, and timezone are given
		 *
		 * @todo
		 * This is currently failling. A query from PHP shows row id 2220957 has "Yaound?" instead of
		 * "Yaoundé", but it's correct in the database itself.
		 */
		 'city-endonym-accents-africa' => array(
			'input' => array(
				'location_name' => 'Yaoundé',
				'locale'        => 'fr_FR',
				'timezone'      => 'Africa/Douala',
			),
			'expected' => array(
				'description' => 'yaoundé',
				'latitude'    => '3.867',
				'longitude'   => '11.517',
				'country'     => 'CM',
			),
		),

		'city-endonym-non-latin-africa' => array(
			'input' => array(
				'location_name' => 'አዲስ አበ',
				'locale'        => 'am',
				'timezone'      => 'Africa/Addis_Ababa',
			),
			'expected' => array(
				'description' => 'addis ababa',
				'latitude'    => '9.025',
				'longitude'   => '38.747',
				'country'     => 'ET',
			),
		),

		'city-endonym-ideographic-asia1' => array(
			'input' => array(
				'location_name' => '白浜町宇佐崎南',
				'locale'        => 'ja',
				'timezone'      => 'Asia/Tokyo',
			),
			'expected' => array(
				'description' => 'shirahamachō-usazakiminami',
				'latitude'    => '34.783',
				'longitude'   => '134.717',
				'country'     => 'JP',
			),
		),

		'city-endonym-ideographic-asia2' => array(
			'input' => array(
				'location_name' => 'تهران',
				'locale'        => 'fa_IR',
				'timezone'      => 'Asia/Tehran',
			),
			'expected' => array(
				'description' => 'tehran',
				'latitude'    => '35.694',
				'longitude'   => '51.422',
				'country'     => 'IR',
			),
		),

		'city-endonym-ideographic-asia3' => array(
			'input' => array(
				'location_name' => 'كراچى',
				'locale'        => 'ur',
				'timezone'      => 'Asia/Karachi',
			),
			'expected' => array(
				'description' => 'karachi',
				'latitude'    => '24.906',
				'longitude'   => '67.082',
				'country'     => 'PK',
			),
		),

		'city-endonym-ideographic-asia4' => array(
			'input' => array(
				'location_name' => '京都',
				'locale'        => 'ja',
				'timezone'      => 'Asia/Tokyo',
			),
			'expected' => array(
				'description' => 'kyoto',
				'latitude'    => '35.021',
				'longitude'   => '135.754',
				'country'     => 'JP',
			),
		),

		'city-endonym-ideographic-asia5' => array(
			'input' => array(
				'location_name' => '東京',
				'locale'        => 'ja',
				'timezone'      => 'Asia/Tokyo',
			),
			'expected' => array(
				'description' => 'tokyo',
				'latitude'    => '35.690',
				'longitude'   => '139.692',
				'country'     => 'JP',
			),
		),

		// The database only has 大阪市 ("Osaka-shi"), not 大阪 ("Osaka"), so an exact match will for 大阪 will fail
		'city-endonym-ideographic-municipal-unit-asia' => array(
			'input' => array(
				'location_name' => '大阪',
				'locale'        => 'ja',
				'timezone'      => 'Asia/Tokyo',
			),
			'expected' => array(
				'description' => 'osaka',
				'latitude'    => '34.694',
				'longitude'   => '135.502',
				'country'     => 'JP',
			),
		),

		'city-endonym-europe' => array(
			'input' => array(
				'location_name' => 'Wien',
				'locale'        => 'de_DE',
				'timezone'      => 'Europe/Berlin',
			),
			'expected' => array(
				'description' => 'vienna',
				'latitude'    => '48.208',
				'longitude'   => '16.372',
				'country'     => 'AT',
			),
		),

		'city-endonym-europe2' => array(
			'input' => array(
				'location_name' => 'Москва',
				'locale'        => 'ru_RU',
				'timezone'      => 'Europe/Moscow',
			),
			'expected' => array(
				'description' => 'moscow',
				'latitude'    => '55.752',
				'longitude'   => '37.616',
				'country'     => 'RU',
			),
		),

		'city-endonym-accents-north-america' => array(
			'input' => array(
				'location_name' => 'Ciudad de México',
				'locale'        => 'en_MX',
				'timezone'      => 'America/Mexico_City',
			),
			'expected' => array(
				'description' => 'mexico city',
				'latitude'    => '19.428',
				'longitude'   => '-99.128',
				'country'     => 'MX',
			),
		),

		'city-endonym-accents-oceania' => array(
			'input' => array(
				'location_name' => 'Hagåtña',
				'locale'        => 'en_US',
				'timezone'      => 'Pacific/Guam',
			),
			'expected' => array(
				'description' => 'hagåtña',
				'latitude'    => '13.476',
				'longitude'   => '144.749',
				'country'     => 'GU',
			),
		),

		'city-endonym-south-america' => array(
			'input' => array(
				'location_name' => 'Bogotá',
				'locale'        => 'es_CO',
				'timezone'      => 'America/Bogota',
			),
			'expected' => array(
				'description' => 'bogotá',
				'latitude'    => '4.610',
				'longitude'   => '-74.082',
				'country'     => 'CO',
			),
		),


		/*
		 * A combination of city, region, and country are given, along with the locale and timezone
		 *
		 * InvalidCity is used in tests that want to bypass the guess_location_from_city() tests and only test the country
		 */
		'1-word-city-region' => array(
			'input' => array(
				'location_name' => 'Portland Maine',
				'locale'        => 'en_US',
				'timezone'      => 'America/New_York',
			),
			'expected' => array(
				'description' => 'portland',
				'latitude'    => '43.661',
				'longitude'   => '-70.255',
				'country'     => 'US',
			),
		),

		'2-word-city-region' => array(
			'input' => array(
				'location_name' => 'São Paulo Brazil',
				'locale'        => 'pt_BR',
				'timezone'      => 'America/Sao_Paulo',
			),
			'expected' => array(
				'description' => 'são paulo',
				'latitude'    => '-23.548',
				'longitude'   => '-46.636',
				'country'     => 'BR',
			),
		),

		'city-1-word-country' => array(
			'input' => array(
				'location_name' => 'InvalidCity Canada',
				'locale'        => 'en_CA',
				'timezone'      => 'America/Vancouver',
			),
			'expected' => array(
				'country' => 'CA',
				'description' => 'canada',
			),
		),

		'city-2-word-country' => array(
			'input' => array(
				'location_name' => 'InvalidCity Dominican Republic',
				'locale'        => 'es_ES',
				'timezone'      => 'America/Santo_Domingo',
			),
			'expected' => array(
				'country' => 'DO',
				'description' => 'dominican republic',
			),
		),

		'city-3-word-country' => array(
			'input' => array(
				'location_name' => 'InvalidCity Central African Republic',
				'locale'        => 'fr_FR',
				'timezone'      => 'Africa/Bangui',
			),
			'expected' => array(
				'country' => 'CF',
				'description' => 'central african republic',
			),
		),

		'country-code' => array(
			'input' => array(
				'location_name' => 'GB',
				'locale'        => 'en_GB',
				'timezone'      => 'Europe/London',
			),
			'expected' => array(
				'country' => 'GB',
				'description' => 'united kingdom',
			),
		),

		'city-country-code' => array(
			'input' => array(
				'location_name' => 'InvalidCity BI',
				'locale'        => 'fr_FR',
				'timezone'      => 'Africa/Bujumbura',
			),
			'expected' => array(
				'country' => 'BI',
				'description' => 'burundi',
			),
		),

		/*
		 * Coordinates should take precedence over IP addresses
		 */
		'coordinates-over-ip-us' => array(
			'input' => array(
				'latitude'  => '47.6062100',
				'longitude' => '-122.3320700',
				'ip'        => '192.0.70.251',  // San Francisco, USA
				'timezone'  => 'America/Los_Angeles',
				'locale'    => 'en_US',
			),
			'expected' => array(
				'description' => false,
				'latitude'    => '47.606',
				'longitude'   => '-122.332',
			),
		),

		'coordinates-over-ip-africa' => array(
			'input' => array(
				'latitude'  => '-19.634233',
				'longitude' => '17.331767',
				'ip'        => '41.190.96.5',   // Tsumeb, Namibia
				'timezone'  => 'Africa/Windhoek',
				'locale'    => 'af',
			),
			'expected' => array(
				'description' => false,
				'latitude'    => '-19.634',
				'longitude'   => '17.332',
			),
		),

		/*
		 * Only the IPv4 address is given
		 *
		 * See https://awebanalysis.com/en/ipv4-directory/
		 */
		'ip-africa' => array(
			'input' => array( 'ip' => '41.191.232.22' ),
			'expected' => array(
				'description' => 'harare',
				'latitude'    => '-17.829',
				'longitude'   => '31.054',
				'country'     => 'ZW',
				'internal'    => true,
			),
		),

		'ip-asia' => array(
			'input' => array( 'ip' => '86.108.55.28' ),
			'expected' => array(
				'description' => 'amman',
				'latitude'    => '31.955',
				'longitude'   => '35.945',
				'country'     => 'JO',
				'internal'    => true,
			),
		),

		'ip-europe' => array(
			'input' => array( 'ip' => '80.95.186.144' ),
			'expected' => array(
				'description' => 'queens island',
				'latitude'    => '54.617',
				'longitude'   => '-5.900',
				'country'     => 'GB',
				'internal'    => true,
			),
		),

		'ip-north-america' => array(
			'input' => array( 'ip' => '189.147.186.0' ),
			'expected' => array(
				'description' => 'ciudad nezahualcoyotl',
				'latitude'    => '19.414',
				'longitude'   => '-99.033',
				'country'     => 'MX',
				'internal'    => true,
			),
		),

		'ip-oceania' => array(
			'input' => array( 'ip' => '116.12.57.122' ),
			'expected' => array(
				'description' => 'auckland',
				'latitude'    => '-36.867',
				'longitude'   => '174.767',
				'country'     => 'NZ',
				'internal'    => true,
			),
		),

		'ip-south-america' => array(
			'input' => array( 'ip' => '181.66.32.136' ),
			'expected' => array(
				'description' => 'lima',
				'latitude'    => '-12.043',
				'longitude'   => '-77.028',
				'country'     => 'PE',
				'internal'    => true,
			),
		),

		/*
		 * Only an IPv6 address is given
		 *
		 * See https://www.google.com/intl/en/ipv6/statistics.html#tab=per-country-ipv6-adoption&tab=per-country-ipv6-adoption
		 * See https://awebanalysis.com/en/ipv6-directory/
		 * See https://www.google.com/search?q=australia+site%3Ahttps%3A%2F%2Fawebanalysis.com%2Fen%2Fipv6-directory%2F
		 */
		'ipv6-africa' => array(
			'input'    => array( 'ip' => '2c0f:f8f0:ffff:ffff:ffff:ffff:ffff:ffff' ),
			'expected' => array(
				'description' => 'harare',
				'latitude'    => '-17.829',
				'longitude'   => '31.054',
				'country'     => 'ZW',
				'internal'    => true,
			),
		),

		'ipv6-asia-anonymized' => array(
			'input'    => array( 'ip' => '2405:200:1000::' ),
			'expected' => array(
				'description' => 'mumbai',
				'latitude'    => '19.014',
				'longitude'   => '72.848',
				'country'     => 'IN',
				'internal'    => true,
			),
		),

		'ipv6-europe-anonymized' => array(
			'input'    => array( 'ip' => '2a02:578:1000::' ),
			'expected' => array(
				'description' => 'sint-niklaas',
				'latitude'    => '51.165',
				'longitude'   => '4.144',
				'country'     => 'BE',
				'internal'    => true,
			),
		),

		'ipv6-north-america-anonymized' => array(
			'input'    => array( 'ip' => '2605:a600::' ),
			'expected' => array(
				'description' => 'mountain view',
				'latitude'    => '37.386',
				'longitude'   => '-122.084',
				'country'     => 'US',
				'internal'    => true,
			),
		),

		'ipv6-oceania-collapsed-prefix' => array(
			'input'    => array( 'ip' => '::ffff:0190:c500' ),
			'expected' => array(
				'description' => 'quedjinup',
				'latitude'    => '-33.634',
				'longitude'   => '115.084',
				'country'     => 'AU',
				'internal'    => true,
			),
		),

		'ipv6-south-america' => array(
			'input'    => array( 'ip' => '2001:1388:6643:2736:10f1:897c:428c:1b3b' ),
			'expected' => array(
				'description' => 'ayacucho',
				'latitude'    => '-13.158',
				'longitude'   => '-74.224',
				'country'     => 'PE',
				'internal'    => true,
			),
		),
	);

	return $cases;
}

run_tests();
